#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#ifdef USE_LSTOOLS
#include <lsapi.h>
#else
#include <old\lsapi.h>
#include "lstoolsemu.h"
#include "lsmmemu.h"
//#include "lsnomm.h"
#endif
#include "..\AggressiveOptimize.h"

#ifdef LS_AUTHORIZATION_UNAVAILABLE
#error Wrong LSAPI.H.  Make sure the path is right.
#endif

//#define SINGLE_WINDOW
//#define NO_SCRIPT_ACTIONS
//#define NO_SKIN

extern const char * const ModuleName;
extern const char * const ModuleVersion;
extern const char * const ModuleAuthor;
extern const char * const ModuleBuild;

extern const char ModuleBangPrefix[];
extern const int ModuleBangPrefixLen;
extern const char ModuleRCPrefix[];
extern const int ModuleRCPrefixLen;
#ifndef SINGLE_WINDOW
extern const char ModuleListKey[];
extern const int ModuleListKeyLen;
#endif


extern lsBangCmdDef Bangs[];

extern
BOOL  (WINAPI *dAlphaBlend)( IN HDC, IN int, IN int, IN int, IN int,
    IN HDC, IN int, IN int, IN int, IN int, IN BLENDFUNCTION);
extern
BOOL (WINAPI *dUpdateLayeredWindow)(HWND,HDC,POINT*,SIZE*,HDC,POINT*,
		COLORREF,BLENDFUNCTION*,DWORD);

#ifndef WS_EX_LAYERED
#define WS_EX_LAYERED           0x00080000
#define ULW_COLORKEY            0x00000001
#define ULW_ALPHA               0x00000002
#define ULW_OPAQUE              0x00000004
#endif

#define UGWL_WNDINFO_PTR 0

#define MSG(n) MessageBox(NULL,n,ModuleName,MB_OK)

#ifndef NO_SCRIPT_ACTIONS
enum ScriptActionTypes
{
  /* valid for windows and module */ 
  OnCreate, OnDestroy,
  OnMouseOver, OnMouseOut,
  OnMouseMove, OnHover, OnUnhover,
  OnLButtonDown, OnLButtonUp,
  OnRButtonDown, OnRButtonUp,
  OnMButtonDown, OnMButtonUp,
  ScriptActionWindowCount,

  /* valid for module only        */ 
  OnInit = ScriptActionWindowCount,
  OnQuit,
  ScriptActionModuleCount,
};
typedef char *ScriptAction;
typedef ScriptAction *ScriptActionSet;
extern const char *ScriptActionStrings[ScriptActionModuleCount+1];
#endif

enum TimerEvents
{
  NotAnEvent = 100,
  HoverTimerEvent
};

struct WndInfo
{
#ifndef SINGLE_WINDOW
  WndInfo *pNext; /* must be first element */ 

  char *Name;
  int Group; /* 0 = no group */ 
#endif

	HWND hWnd;
	HWND hParentWnd;

	POINT Pos;
	SIZE Size;
	LSMONITOR Monitor;
	RECT rMonitor;

	LONG Alpha;

	bool bTrueAlpha;
	bool bOnTop;
	bool bHidden;

	HBITMAP hbmBackBuffer;

#ifndef NO_SKIN
  lsEdgeMap emSkin;
#endif

#ifndef NO_SCRIPT_ACTIONS
  UINT_PTR HoverTimerID;
  POINT HoverPoint;
  bool bHovering;
  bool bMouseOver;
#endif

#if !defined(NO_SCRIPT_ACTIONS) && !defined(SINGLE_WINDOW)
  /* each window can have its own actions  */ 
  /* if it does, the module one is ignored */ 
  /* length = ScriptActionWindowCount      */ 
  ScriptActionSet Actions;

  int HoverTime; /* 0 = use global hover time */ 
#endif
};

/* Global data for the module.   */ 
/* Grouped together partially so */ 
/* we can clear it all at once.  */ 
struct ModInfo
{
	HWND hBottomParentWnd;
	HINSTANCE hInstance;

#ifdef SINGLE_WINDOW
	bool bWaitForBox;
  WndInfo MainWnd;
#else
  WndInfo *pWnds;
#endif

#ifndef NO_SCRIPT_ACTIONS
  /* global module-wide actions       */ 
  /* length = ScriptActionModuleCount */ 
  ScriptActionSet Actions;

  int HoverTime;
#endif
};

extern ModInfo Module; /* global module data */ 

struct Config
{
	char key[256];
	int keylen;

	char one[MAX_LINE_LENGTH],two[MAX_LINE_LENGTH],
		three[MAX_LINE_LENGTH],four[MAX_LINE_LENGTH];
	char *tokens[4];
	char line[MAX_LINE_LENGTH];
};

class ConfigState
{
public:
	int keylen;
	Config *config;

	ConfigState(Config *c,const char *key,int len) : config(c),keylen(len)
		{ lstrcpyn(config->key+config->keylen,key,len+1); config->keylen+=keylen; }
	~ConfigState()
		{ config->keylen-=keylen; config->key[config->keylen]=0; }
};

// read configuration for the whole module
void ReadConfigMod(ModInfo *m,Config *c,const char *key,int keylen);
void FreeConfigMod(ModInfo *m);
// read configuration for a window
void ReadConfigWnd(WndInfo *w,Config *c,const char *key,int keylen);
void FreeConfigWnd(WndInfo *w);

// create a window and mark it as part of litestep
bool CreateWnd(WndInfo *w);
// update size and position of window
void UpdateWndPos(WndInfo *w);
// update on top/bottem setting
void UpdateWndOnTop(WndInfo *w);
// update hide/show
void UpdateWndHidden(WndInfo *w);

// paint the window
void PaintWnd(WndInfo *w,HDC hdc);
void RedrawWnd(WndInfo *w);

#ifndef NO_SCRIPT_ACTIONS
// this needs to go into lstools
void ReplaceAtVars(char *dest,int maxchars,const char *src,
                   int (*LookupVar)(char *dest,int maxchars,const char *src,LPARAM lParam),
                   LPARAM lParam);

void ExecActionString(HWND hCaller,WndInfo *w,const char *sAction);
void ExecAction(WndInfo *w,int iAction);

#ifndef SINGLE_WINDOW
inline int GetHoverTime(WndInfo *w) { return w->HoverTime?w->HoverTime:Module.HoverTime; }
#else /* SINGLE_WINDOW */ 
inline int GetHoverTime(WndInfo *) { return Module.HoverTime; }
#endif /* SINGLE_WINDOW */ 

inline bool IsKeyDown(int key) { return (GetKeyState(key)&0x8000)!=0; }

#endif /* NO_SCRIPT_ACTIONS */ 

#ifndef SINGLE_WINDOW
// create all of the windows
void CreateWindows(void);
// read a windows config
int ParseBoxProc(void *elem,const char *line,DWORD conf);
// close and release the windows
void FreeWndProc(void *elem,DWORD);
void AddWindow(WndInfo *w);
#endif


inline int atoi(const char *ptr)
{ int i=0,s=(*ptr=='-')?(++ptr,-1):1; while (*ptr>='0' && *ptr<='9') i = i*10 + *ptr++ - '0'; return i*s; }

