#include "module.h"

#ifndef NO_SCRIPT_ACTIONS
static void UpdateHover(WndInfo *w,bool bOnMouseMove)
{
  bool bHoveringNow,bMouseOverNow;
  POINT pt;
  RECT r;

  ASSERT(READABLE(w,sizeof(*w)));
  ASSERT(WRITEABLE(w,sizeof(*w)));

  GetCursorPos(&pt);
  GetWindowRect(w->hWnd,&r);

  bMouseOverNow = PtInRect(&r,pt)!=0;
  bHoveringNow = bMouseOverNow && pt.x==w->HoverPoint.x && pt.y==w->HoverPoint.y;

  if (bOnMouseMove)
  {
    if (bMouseOverNow && !w->bMouseOver)
    {
      w->bMouseOver = true;
      ExecAction(w,OnMouseOver);
      SetCapture(w->hWnd);
    }
    if (w->bHovering && !bHoveringNow)
      { w->bHovering = false; ExecAction(w,OnUnhover); }
    w->HoverPoint = pt;
    w->HoverTimerID = 
      SetTimer(w->hWnd,HoverTimerEvent,GetHoverTime(w),NULL);
  }
  else
  {
    if (bHoveringNow && !w->bHovering)
      { w->bHovering = true; ExecAction(w,OnHover); }
  }
  if (w->bMouseOver && !bMouseOverNow)
  {
    w->bMouseOver = false; ExecAction(w,OnMouseOut);
    ReleaseCapture();
    KillTimer(w->hWnd,w->HoverTimerID);
  }
}
#endif /* NO_SCRIPT_ACTIONS */ 


LRESULT CALLBACK WndProc(HWND hwnd, UINT message, WPARAM wParam, LPARAM lParam)
{
	WndInfo *w = (WndInfo *)GetWindowLong(hwnd,UGWL_WNDINFO_PTR);
	switch (message) {
	case WM_SHOWWINDOW:
		{
			HDC hdcScreen = GetDC(NULL);

			if (w->hbmBackBuffer==NULL)
				w->hbmBackBuffer = CreateCompatibleBitmap(hdcScreen,w->Size.cx,w->Size.cy);

			ReleaseDC(NULL,hdcScreen);

			RedrawWnd(w);
		}
		break;

	case WM_DESTROY:
		{
     #ifndef NO_SCRIPT_ACTIONS
      /* OnCreate is in CreateWnd() */ 
      ExecAction(w,OnDestroy);
     #endif

			if (w->hbmBackBuffer!=NULL) DeleteObject(w->hbmBackBuffer);
		}
		break;

	case WM_PAINT:
		{
      /* not used for layered windows */ 

			PAINTSTRUCT ps;
			HDC hdc = BeginPaint(hwnd,&ps);
			HDC hdcWnd = CreateCompatibleDC(hdc);
			HBITMAP hbmOld = (HBITMAP)SelectObject(hdcWnd,w->hbmBackBuffer);

			BitBlt(hdc,0,0,w->Size.cx,w->Size.cy,hdcWnd,0,0,SRCCOPY);

			SelectObject(hdcWnd,hbmOld);
			DeleteDC(hdcWnd);

			EndPaint(hwnd,&ps);
		}
		break;

#ifndef NO_SCRIPT_ACTIONS
  case WM_LBUTTONDOWN: ExecAction(w,OnLButtonDown); break;
  case WM_LBUTTONUP:   ExecAction(w,OnLButtonUp  ); break;
  case WM_RBUTTONDOWN: ExecAction(w,OnRButtonDown); break;
  case WM_RBUTTONUP:   ExecAction(w,OnRButtonUp  ); break;
  case WM_MBUTTONDOWN: ExecAction(w,OnMButtonDown); break;
  case WM_MBUTTONUP:   ExecAction(w,OnMButtonUp  ); break;
  case WM_MOUSEMOVE:   UpdateHover(w,true);
                       ExecAction(w,OnMouseMove  ); break;

  case WM_TIMER:
    if (wParam == HoverTimerEvent)
  case WM_CAPTURECHANGED:
      UpdateHover(w,false);
    break;
#endif

	case LM_GETREVID:
		{
  		PTSTR buffer = (PTSTR)lParam;

			if (wParam == 0)
			{
				lstrcpy(buffer, ModuleName);
				lstrcat(buffer, ".dll: ");
				lstrcat(buffer, ModuleVersion);
				lstrcat(buffer, " (");
				lstrcat(buffer, ModuleAuthor);
				lstrcat(buffer, ")");
			}
			else
			{
				*buffer = 0;
			}
			return lstrlen(buffer);
		}

	default:
		return DefWindowProc(hwnd,message,wParam,lParam);
	}

	return 0;
}